from airtouch2.protocol.messages.ResponseMessage import ResponseMessage
from airtouch2.protocol.messages.RequestState import RequestState
from airtouch2.protocol.messages.ac_commands import ChangeSetTemperature, ToggleAC, SetFanSpeed, SetMode
from airtouch2.protocol.messages.group_commands import ToggleGroup, ChangeDamper


from __future__ import annotations
import logging
from itertools import compress
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from airtouch2.AT2Client import AT2Client
from airtouch2.protocol.enums import ACFanSpeedReference, ACBrand, ACMode
from airtouch2.protocol.messages import ChangeSetTemperature, ResponseMessage, SetFanSpeed, SetMode, ToggleAC
from airtouch2.protocol.lookups import GATEWAYID_BRAND_LOOKUP

OPEN_ISSUE_TEXT = "please open an issue and detail your system to me:\n\thttps://github.com/nathanvdh/airtouch2-python/issues/new"

_LOGGER = logging.getLogger(__name__)
class AT2Aircon:
    def __init__(self, number: int, client: AT2Client, response_message: ResponseMessage):
        self.number: int = number
        self._client: AT2Client = client
        if response_message:
            self.update(response_message)
        else:
            raise ValueError("Null response message provided")

    # TODO: read through app source code more and do this more properly
    # Currently I'm assuming:
    #   4 speed is always LOW, MED, HIGH, POWERFUL (EXCEPT for fujitsus with 4 speeds)
    #   3 speed is always LOW, MED, HIGH
    #   2 speed is always LOW, HIGH
    #   Daikins have no Auto
    #   Gateway ID 0x14 has no Auto
    #   Gateway IDs 0xFF with 3 speed have no Auto
    # These are based on the app decompiled code
    def _set_supported_fan_speeds(self, num_supported_speeds: int, gateway_id: int):
        all_speeds:list[ACFanSpeedReference] = list(ACFanSpeedReference.__members__.values())
        if self.brand == ACBrand.FUJITSU and num_supported_speeds == 4:
            self.supported_fan_speeds = all_speeds[:5]
            return
        # Check cases that don't support Auto mode
        if (self.brand == ACBrand.DAIKIN or (gateway_id == 0xFF and num_supported_speeds == 3) or gateway_id == 0x14):
            self.supported_fan_speeds = []
        else:
            self.supported_fan_speeds = [ACFanSpeedReference.AUTO]

        if num_supported_speeds > 2:
            self.supported_fan_speeds += all_speeds[2:2+num_supported_speeds]
        elif num_supported_speeds == 2:
            self.supported_fan_speeds += [ACFanSpeedReference.LOW, ACFanSpeedReference.HIGH]
        elif num_supported_speeds < 2:
            _LOGGER.warning(f"AC{self.number} reports less than 2 supported fan speeds, this is unusual - " + OPEN_ISSUE_TEXT)

    def _set_fan_speed_from_val(self, speed_val: int):
        if speed_val < 5:
            # Units with no Auto speed still start with low = 1
            if ACFanSpeedReference.AUTO not in self.supported_fan_speeds:
                speed_val -= 1
            self.fan_speed = self.supported_fan_speeds[speed_val]
        else:
            self.fan_speed = ACFanSpeedReference.AUTO

    def _get_speed_val_from_speed(self, fan_speed: ACFanSpeedReference):
        speed_val = self.supported_fan_speeds.index(fan_speed)
        # Units with no Auto speed still start with low = 1
        if ACFanSpeedReference.AUTO not in self.supported_fan_speeds:
            speed_val += 1
        return speed_val

    def update(self, response_message: ResponseMessage) -> None:
        self.system_name = response_message.system_name

        # Flags
        self.on = response_message.ac_active[self.number]
        self.safety = response_message.ac_safety[self.number]
        self.error = response_message.ac_error[self.number]
        self.turbo = response_message.ac_turbo[self.number]
        self.spill = response_message.ac_spill[self.number]

        # Error code
        self.error_code = response_message.ac_error_code[self.number]

        # Temperatures
        self.measured_temp = response_message.ac_measured_temp[self.number]
        if (self.measured_temp <= 0):
            self.measured_temp = response_message.touchpad_temp
        self.set_temp = response_message.ac_set_temp[self.number]

        # Brand
        brand = response_message.ac_brand[self.number]
        gateway_id = response_message.ac_gateway_id[self.number]
        # Brand based on gateway ID takes priority
        if gateway_id:
            if gateway_id not in GATEWAYID_BRAND_LOOKUP:
                _LOGGER.warning(f"AC{self.number} has an unfamiliar gateway ID: {hex(gateway_id)} - " + OPEN_ISSUE_TEXT + "\nInclude the gateway ID shown above")
                self.brand = brand
            else:
                self.brand = GATEWAYID_BRAND_LOOKUP[gateway_id]
        else:
            self.brand = brand

        # Modes
        self.mode = response_message.ac_mode[self.number]
        num_fan_speeds = response_message.ac_num_fan_speeds[self.number]
        fan_speed_val = response_message.ac_fan_speed[self.number]
        self._set_supported_fan_speeds(num_fan_speeds, gateway_id)
        self._set_fan_speed_from_val(fan_speed_val)

        # TODO: Handle this?
        if brand == ACBrand.NONE and gateway_id == 0:
            _LOGGER.warning(f"AC{self.number} appears disconnected from airtouch")

        self.name = response_message.ac_name[self.number]

    async def inc_dec_set_temp(self, inc: bool):
        await self._client.send_command(ChangeSetTemperature(self.number, inc))

    async def set_set_temp(self, new_temp: int):
        temp_diff = new_temp - self.set_temp
        inc = temp_diff > 0
        for i in range(abs(temp_diff)):
            await self.inc_dec_set_temp(inc)

    async def _turn_on_off(self, on: bool):
        if self.on != on:
            await self._client.send_command(ToggleAC(self.number))

    async def turn_off(self):
        await self._turn_on_off(False)

    async def turn_on(self):
        await self._turn_on_off(True)

    async def set_fan_speed(self, fan_speed: ACFanSpeedReference):
        if fan_speed in self.supported_fan_speeds:
            await self._client.send_command(SetFanSpeed(self.number, self._get_speed_val_from_speed(fan_speed)))
        else:
            _LOGGER.warning(f"Cannot set fan speed to unsupported value {fan_speed}")

    async def set_mode(self, mode: ACMode):
        await self._client.send_command(SetMode(self.number, mode))

    def get_status_strings(self):
        flags = [self.error, self.safety, self.spill, self.turbo]
        flag_names = ['ERROR', 'SAFETY', 'SPILL', 'TURBO']
        statuses = list(compress(flag_names, flags))
        if not statuses:
            statuses.append('NORMAL')
        return statuses

    def __str__(self):
        return f"""
        System Name:\t\t{self.system_name}
        AC Number:\t\t{self.number}
        AC Name:\t\t{self.name}
        On:\t\t\t{self.on}
        Status:\t\t\t{self.get_status_strings()}
        Error Code:\t\t{self.error_code}
        Mode:\t\t\t{self.mode}
        Fan Speed:\t\t{self.fan_speed}
        Supported Speeds:\t{[s.name for s in self.supported_fan_speeds]}
        Measured Temp:\t\t{self.measured_temp}
        Set Temp:\t\t{self.set_temp}
        Brand:\t\t\t{self.brand}
        """



from airtouch2.protocol.messages.Message import Message
from airtouch2.protocol.constants import CommandMessageConstants, MessageLength
from abc import abstractmethod
class CommandMessage(Message):
    """ Command message base class from which all airtouch2 command messages are derived"""
    length: MessageLength = MessageLength.COMMAND

    # command messages are only sent so only need to be serialized
    def _serialize(self, prefilled_msg: bytearray) -> bytearray:
        prefilled_msg[0] = CommandMessageConstants.BYTE_0
        prefilled_msg[2] = CommandMessageConstants.BYTE_2
        return self.add_checksum(prefilled_msg)

    @abstractmethod
    def serialize(self) -> bytearray:
        pass


from abc import ABC
from airtouch2.protocol.constants import MessageLength

class Message(ABC):
    """ Message base class from which all airtouch 2 communications messages are derived"""
    length: MessageLength = MessageLength.UNDETERMINED

    def add_checksum(self, serial_msg: bytearray) -> bytearray:
        serial_msg[self.length-1] = self.checksum(serial_msg)
        return serial_msg

    @staticmethod
    def checksum(serial_msg: bytearray) -> int:
        sum: int = 0
        for b in serial_msg:
            sum += b
        sum = (sum % 256)
        return sum


import asyncio
from datetime import datetime
import errno
import logging
from socket import gaierror
from typing import Callable

from airtouch2.protocol.constants import MessageLength
from airtouch2.protocol.messages import RequestState, ResponseMessage
from airtouch2.AT2Aircon import AT2Aircon
from airtouch2.AT2Group import AT2Group
from airtouch2.protocol.messages.CommandMessage import CommandMessage

_LOGGER = logging.getLogger(__name__)

NetworkOrHostDownErrors = (errno.EHOSTUNREACH, errno.ECONNREFUSED,  errno.ETIMEDOUT,
                           errno.ENETDOWN, errno.ENETUNREACH, errno.ENETRESET, errno.ECONNABORTED)


class AT2Client:
    def __init__(self, host: str, dump: bool = False) -> None:
        self._host_ip: str = host
        self._host_port: int = 8899
        self._reader: asyncio.StreamReader | None = None
        self._writer: asyncio.StreamWriter | None = None
        self._dump: bool = dump
        self._stop: bool = True
        self._read_task: asyncio.Task[None] | None = None
        self._got_response: asyncio.Event = asyncio.Event()
        self._callbacks: list[Callable] = []
        self.aircons: list[AT2Aircon] = []
        self.groups: list[AT2Group] = []
        self.system_name: str = "UNKNOWN"
        self.active: bool = False

    async def connect(self) -> bool:
        """Opens connection to the server, returns True/False if successful/unsuccessful"""
        _LOGGER.debug(
            f'Connecting to {self._host_ip} on port {self._host_port}')
        try:
            self._reader, self._writer = await asyncio.open_connection(self._host_ip, self._host_port)
        except OSError as e:
            _LOGGER.warning(f"Could not connect to host {self._host_ip}")
            if isinstance(e, gaierror):
                # provided ip or port is rubbish/invalid
                pass
            elif e.errno not in NetworkOrHostDownErrors:
                raise e
            return False
        else:
            return True

    async def run(self, create_task: Callable = asyncio.create_task) -> None:
        """Starts the processing of incoming information from the server, using the provided create_task function"""
        _LOGGER.debug("Starting listener task")
        self._stop = False
        self._read_task = create_task(self._listen_for_updates())
        await self.send_command(RequestState())
        await asyncio.wait_for(self._got_response.wait(), timeout=5)

    async def stop(self) -> None:
        if not self._read_task:
            raise RuntimeError("Client task is not running")
        self._stop = True
        self._read_task.cancel()
        try:
            await self._read_task
        except asyncio.CancelledError as e:
            # Eat the expected exception
            pass
        self._got_response.clear()

    async def send_command(self, command: CommandMessage) -> None:
        if not self._writer:
            raise RuntimeError("Client is not connected - call run() first")
        else:
            _LOGGER.debug(f"Sending {command.__class__.__name__}")
            self._writer.write(command.serialize())
            self._got_response.clear()
            await self._writer.drain()
            await self._got_response.wait()

    def add_callback(self, func: Callable):
        self._callbacks.append(func)

        def remove_callback() -> None:
            if func in self._callbacks:
                self._callbacks.remove(func)

        return remove_callback

    async def _read_response(self) -> ResponseMessage:
        if not (self._reader and self._writer):
            raise RuntimeError(
                "Client is not connected - call connect() first")
        _LOGGER.debug("Waiting for response")
        resp = await self._reader.read(MessageLength.RESPONSE)
        # TODO: handle socket exceptions?
        _LOGGER.debug("Got response")
        while len(resp) != MessageLength.RESPONSE:
            if not resp:
                _LOGGER.debug("Connection lost, reconnecting")
                self._got_response.clear()
                await self._try_reconnect()
            else:
                _LOGGER.debug(
                    f"Invalid length message received ({len(resp)}) - ignoring")
            resp = await self._reader.read(MessageLength.RESPONSE)
        if self._dump:
            # blocks but is only used for dev and debugging
            with open('response_' + datetime.now().strftime("%m-%d-%Y_%H-%M-%S") + '.dump', 'wb') as f:
                f.write(resp)

        return ResponseMessage(resp)

    async def _try_reconnect(self) -> None:
        retries = 0
        while not await self.connect():
            await asyncio.sleep(0.001 * (10**retries) if retries < 4 else 10)
            retries += 1
            if not retries % 50 or retries == 4:
                _LOGGER.debug(
                    "Server is not responding, will continue trying to reconnect every 10s")
        # reconnected
        await self.send_command(RequestState())

    async def _listen_for_updates(self) -> None:
        while not self._stop:
            try:
                resp = await self._read_response()
            except asyncio.CancelledError as e:
                raise e
            else:
                # ACs
                if not self.aircons:
                    self.aircons.append(AT2Aircon(0, self, resp))
                    _LOGGER.debug(self.aircons[0])
                else:
                    for aircon in self.aircons:
                        aircon.update(resp)
                        _LOGGER.debug(aircon)
                # Groups
                if not self.groups or len(self.groups) != resp.num_groups:
                    self.groups.clear()
                    for i in range(resp.num_groups):
                        await asyncio.sleep(0)
                        self.groups.append(AT2Group(self, i, resp))
                        _LOGGER.debug(self.groups[i])
                else:
                    for group in self.groups:
                        await asyncio.sleep(0)
                        group.update(resp)
                        _LOGGER.debug(group)
                self._got_response.set()
                for func in self._callbacks:
                    await asyncio.sleep(0)
                    func()



from enum import IntEnum

class ACMode(IntEnum):
    AUTO = 0
    HEAT = 1
    DRY = 2
    FAN = 3
    COOL = 4

    def __str__(self):
        return self._name_

class ACFanSpeedReference(IntEnum):
    AUTO = 0
    QUIET = 1
    LOW = 2
    MEDIUM = 3
    HIGH = 4
    POWERFUL = 5

    def __str__(self):
        return self._name_

# From adding AC2 on the panel and cycling through brands
class ACBrand(IntEnum):
    NONE = 0
    DAIKIN = 1
    FUJITSU = 2
    HITACHI = 3
    LG = 4
    MITSUBISHI_ELECTRIC = 5
    MITSUBISHI_HEAVY_IND = 6
    PANASONIC = 7
    SAMSUNG = 8
    TOSHIBA = 9

    def __str__(self):
        return self._name_



from airtouch2.protocol.messages.CommandMessage import CommandMessage
from airtouch2.protocol.constants import CommandMessageType

class RequestState(CommandMessage):
    """ Command to request the state of the airtouch 2 system"""

    def serialize(self) -> bytearray:
        serial_msg: bytearray = bytearray(self.length)
        serial_msg[1] = CommandMessageType.REQUEST_STATE
        return super()._serialize(serial_msg)